/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.command;

import static de.ozgcloud.vorgang.command.PersistedCommand.*;

import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.stream.Stream;

import org.apache.commons.collections4.MapUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.stereotype.Service;
import org.springframework.validation.annotation.Validated;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.callcontext.CallContext;
import de.ozgcloud.vorgang.callcontext.CallContextUser;
import de.ozgcloud.vorgang.callcontext.CurrentUserService;
import de.ozgcloud.vorgang.callcontext.User;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.common.errorhandling.RevokeFailedException;
import de.ozgcloud.vorgang.vorgang.VorgangService;
import lombok.NonNull;

@Service
@Validated
public class CommandService {

	@Autowired
	private CommandRepository repository;
	@Autowired
	private ApplicationEventPublisher publisher;

	@Autowired
	private VorgangService vorgangService;
	@Autowired
	private CurrentUserService userService;

	public Command createCommand(CreateCommandRequest request) {
		validateVorgang(request.getVorgangId(), request.getOrder());
		var persisted = saveCommand(buildCommand(request));
		publisher.publishEvent(new CommandCreatedEvent(persisted));
		return persisted;
	}

	void validateVorgang(String vorgangId, String order) {
		if (!Order.UNLOCK_VORGANG.isMeant(order) && vorgangService.isVorgangLocked(vorgangId)) {
			throw new TechnicalException("Cannot create command. Vorgang %s is locked.".formatted(vorgangId));
		}
	}

	private Command buildCommand(CreateCommandRequest request) {
		var builder = PersistedCommand.builder()
				.id(UUID.randomUUID().toString())
				.vorgangId(request.getVorgangId())
				.order(request.getOrder())
				.status(CommandStatus.PENDING)
				.createdAt(ZonedDateTime.now())
				.relationId(request.getRelationId())
				.relationVersion(request.getRelationVersion())
				.redirectRequest(request.getRedirectRequest())
				.body(request.getBody())
				.bodyObject(request.getBodyObject());

		userService.findUser().map(CallContextUser::getClientName).ifPresent(builder::createdByClientName);
		builder = addUserIfExists(builder, Optional.ofNullable(request.getCallContext()).map(CallContext::getUser), userService.findUser());
		return builder.build();
	}

	private PersistedCommand.PersistedCommandBuilder addUserIfExists(PersistedCommand.PersistedCommandBuilder builder, Optional<User> user,
			Optional<CallContextUser> contextUser) {
		if (contextUser.isPresent()) {
			return builder.createdBy(contextUser.get().getUserId().orElse(null)).createdByName(contextUser.get().getUserName().orElse(null));
		}

		return user.map(u -> builder.createdBy(u.getId()).createdByName(u.getName())).orElse(builder);
	}

	Command saveCommand(Command command) {
		repository.save(command);

		return command;
	}

	public Stream<Command> findFinishedVorgangLoeschenCommandsOlderThen(@NonNull ZonedDateTime createdAfter) {
		return repository.findCommands(Order.VORGANG_LOESCHEN, CommandStatus.FINISHED, createdAfter);
	}

	public void setCommandFinished(String commandId) {
		setCommandFinished(commandId, null);
	}

	public void setCommandFinished(String commandId, String createdResource) {
		var command = getById(commandId);
		if (shouldRevoke(command)) {
			command = repository.setRevokeStatus(commandId, createdResource);
			publishRevokeCommandEvent(command);
			return;
		}
		Optional.ofNullable(createdResource).ifPresentOrElse(resource -> repository.finishCommand(commandId, resource),
				() -> repository.finishCommand(commandId));
		getCompletableParentId(command).ifPresent(this::publishCommandExecutedEvent);
	}

	boolean shouldRevoke(Command command) {
		return command.getStatus() == CommandStatus.REVOKE_PENDING || isParentCommandFailed(command);
	}

	boolean isParentCommandFailed(Command command) {
		return getParentId(command).filter(repository::isCommandFailed).isPresent();
	}

	Optional<String> getCompletableParentId(Command command) {
		return getParentId(command).filter(repository::isCompleteIfSubsCompleted)
				.filter(parentId -> !repository.existsNotFinishedSubCommands(parentId));
	}

	Optional<String> getParentId(Command command) {
		return Optional.ofNullable(command.getBodyObject()).map(body -> MapUtils.getString(body, PROPERTY_PARENT_ID));
	}

	public void publishCommandExecutedEvent(String commandId) {
		findById(commandId).ifPresent(command -> publisher.publishEvent(new CommandExecutedEvent(command)));
	}

	public void setCommandError(String commandId, String errorMessage) {
		repository.setErrorMessage(commandId, errorMessage);
		repository.getParentId(commandId).map(this::setErrorStatus).filter(this::notErrorStatus)
				.map(Command::getId).ifPresent(parentId -> handleCommandError(commandId, parentId));
	}

	Command setErrorStatus(String commandId) {
		return Optional.ofNullable(repository.updateCommandStatusAndReturnPrevious(commandId, CommandStatus.ERROR))
				.orElseThrow(() -> new NotFoundException(Command.class, commandId));
	}

	boolean notErrorStatus(Command command) {
		return command.getStatus() != CommandStatus.ERROR;
	}

	void handleCommandError(String commandId, String parentId) {
		publishCommandFailedEvent(parentId, "Command %s failed because subcommand %s failed".formatted(parentId, commandId));
		revokeSubCommands(parentId);
	}

	void publishCommandFailedEvent(String commandId, String errorMessage) {
		publisher.publishEvent(new CommandFailedEvent(commandId, errorMessage));
	}

	void revokeSubCommands(String parentId) {
		repository.findSubCommandIds(parentId).forEach(this::revokeCommandSilent);
	}

	public void setCommandRevoked(String commandId) {
		repository.updateCommandStatus(commandId, CommandStatus.REVOKED);
	}

	public Command revokeCommand(String commandId) {
		var updatedCommand = revokeCommandSilent(commandId).orElseThrow(() -> new NotFoundException(Command.class, commandId));
		if (!isRevokePending(updatedCommand)) {
			throw new RevokeFailedException(commandId, "Unexpected status '%s'. (expected: REVOKE_PENDING)".formatted(updatedCommand.getStatus()));
		}
		return updatedCommand;
	}

	Optional<Command> revokeCommandSilent(String commandId) {
		var updatedCommand = repository.setRevokeStatusIfNotPending(commandId);
		updatedCommand.filter(this::isRevokePending).ifPresent(this::publishRevokeCommandEvent);
		return updatedCommand;
	}

	boolean isRevokePending(Command command) {
		return command.getStatus() == CommandStatus.REVOKE_PENDING;
	}

	void publishRevokeCommandEvent(Command command) {
		publisher.publishEvent(new RevokeCommandEvent(command));
	}

	public boolean existsPendingCommands(String vorgangId) {
		return repository.existsPendingCommands(vorgangId);
	}

	public List<Command> getPendingCommands(String relationId) {
		return repository.getPendingCommands(relationId);
	}

	public Stream<Command> findCommands(String vorgangId, Collection<CommandStatus> status, Optional<Order> order) {
		return repository.findCommands(vorgangId, status, order).stream();
	}

	public void setPreviousState(String commandId, Map<String, Object> previousState) {
		repository.patch(commandId, previousState);
	}

	public void deleteAllByVorgang(String vorgangId) {
		repository.deleteAllByVorgang(vorgangId);
	}

	public boolean existsCommandWithUserId(String userId) {
		return repository.existsCommandWithUserId(userId);
	}

	public Stream<Command> createSubCommands(CreateSubCommandsRequest request) {
		var parentCommand = getPendingCommand(request.getParentId());
		repository.addSubCommands(parentCommand.getId(), buildSubCommandValues(request));
		return request.getSubCommands().stream().map(subCommand -> buildCreateCommandRequest(subCommand, parentCommand))
				.map(this::createCommand);
	}

	Command getPendingCommand(String commandId) {
		var command = getById(commandId);
		if (command.getStatus() != CommandStatus.PENDING) {
			throw new TechnicalException("Cannot add sub commands to command '%s'. Parent command must have status PENDING".formatted(commandId));
		}
		return command;
	}

	public Command getById(String commandId) {
		return findById(commandId).orElseThrow(() -> new NotFoundException(Command.class, commandId));
	}

	public Optional<Command> findById(String commandId) {
		return repository.findById(commandId);
	}

	Map<String, Object> buildSubCommandValues(CreateSubCommandsRequest request) {
		var executionMode = request.getExecutionMode();
		return Map.of(
				PROPERTY_EXECUTION_MODE, executionMode.name(),
				PROPERTY_COMPLETE_IF_SUBS_COMPLETED, request.isCompletedIfSubsCompleted());
	}

	CreateCommandRequest buildCreateCommandRequest(CreateSubCommandRequest createCommand, Command parentCommand) {
		return CreateCommandRequest.builder()
				.vorgangId(parentCommand.getVorgangId())
				.relationId(createCommand.getRelationId())
				.relationVersion(createCommand.getRelationVersion())
				.order(createCommand.getOrder())
				.bodyObject(addParentIdToBodyObject(createCommand.getBodyObj(), parentCommand.getId()))
				.build();
	}

	Map<String, Object> addParentIdToBodyObject(Map<String, Object> bodyObject, String parentId) {
		var bodyObjectMutable = new HashMap<>(bodyObject);
		bodyObjectMutable.put(PROPERTY_PARENT_ID, parentId);
		return Collections.unmodifiableMap(bodyObjectMutable);
	}

}