/*
 * Copyright (C) 2022-2023 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Stream;

import jakarta.validation.Valid;

import org.apache.commons.collections.MapUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.commons.lang3.tuple.Pair;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.security.access.AccessDeniedException;
import org.springframework.stereotype.Service;
import org.springframework.validation.annotation.Validated;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.VorgangAssignedEvent;
import de.ozgcloud.command.VorgangCreatedEvent;
import de.ozgcloud.command.VorgangLockedEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeMap;
import de.ozgcloud.vorgang.clientattribute.ClientAttributeReadPermitted;
import de.ozgcloud.vorgang.clientattribute.ClientAttributesMap;
import de.ozgcloud.vorgang.common.errorhandling.NotFoundException;
import de.ozgcloud.vorgang.servicekonto.ServiceKonto;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;

@Validated
@Service
@RequiredArgsConstructor
public class VorgangService {

	private static final String BODY_ASSIGNED_TO_FIELD = "assignedTo";
	public static final String BODY_OBJECT_AKTENZEICHEN = "aktenzeichen";
	static final String KEY_HEADER_LOCK = "%s.%s".formatted(Vorgang.MONGODB_FIELDNAME_HEADER, VorgangHead.FIELD_LOCK);

	private final VorgangAuthorizationService vorgangAuthenticationService;
	private final VorgangRepository repository;

	private final ApplicationEventPublisher publisher;
	private final AktenzeichenProvider aktenzeichenProvider;

	private final ClientAttributeReadPermitted readIsPermitted;

	private final LabelProcessor kopControlDataMapper;

	private final VorgangStubMapper stubMapper;

	public Vorgang startCreation(@Valid Eingang eingang) {
		var mappedEingang = kopControlDataMapper.moveLabelsToControlData(eingang);

		return repository.save(buildVorgang(mappedEingang, true));
	}

	private Vorgang buildVorgang(Eingang eingang, boolean inCreation) {
		return Vorgang.builder()
				.name(eingang.getHeader().getFormName())
				.aktenzeichen(aktenzeichenProvider.provideAktenzeichen(eingang))
				.nummer(getVorgangNummer(eingang.getHeader()))
				.formEngineName(eingang.getHeader().getFormEngineName())
				.inCreation(inCreation)
				.header(buildVorgangHead(eingang))
				.eingang(eingang)
				.build();
	}

	String getVorgangNummer(@NonNull EingangHeader header) {
		return Optional.ofNullable(header.getVorgangNummer()).filter(StringUtils::isNotBlank)
				.orElseGet(header::getRequestId);
	}

	private Optional<ServiceKonto> getServiceKonto(Eingang eingang) {
		return Optional.ofNullable(eingang.getHeader().getServiceKonto());
	}

	private VorgangHead buildVorgangHead(Eingang eingang) {
		var vorgangHeadBuilder = VorgangHead.builder();

		getServiceKonto(eingang).ifPresent(vorgangHeadBuilder::serviceKonto);

		return vorgangHeadBuilder.build();
	}

	public void finishCreation(String vorgangId, List<IncomingFileGroup> attachments, List<IncomingFile> representations) {
		// TODO nicht den ganzen Vorgang laden und überschreiben sondern nur die
		// notwendigen Felder anpassen (patch)
		var loaded = repository.findById(vorgangId)
				.orElseThrow(() -> new IllegalStateException("Cannot load vorgang for finishing creation -" + vorgangId));
		var vorgang = addAttachmentsAndRepresentations(loaded, attachments, representations).inCreation(false).build();

		var savedVorgang = repository.save(vorgang);

		publishVorgangeCreatedEvent(savedVorgang);
	}

	private Vorgang.VorgangBuilder addAttachmentsAndRepresentations(Vorgang vorgang, List<IncomingFileGroup> attachments,
			List<IncomingFile> representations) {
		var eingang = vorgang.getEingangs().get(0).toBuilder()
				.attachments(attachments)
				.representations(representations)
				.build();

		return vorgang.toBuilder().clearEingangs().eingang(eingang);
	}

	private void publishVorgangeCreatedEvent(Vorgang vorgang) {
		publisher.publishEvent(new VorgangCreatedEvent(vorgang.getId()));
	}

	public Vorgang getById(String vorgangId, FilterCriteria filterCriteria) {
		var vorgang = getById(vorgangId);

		if (notAuthorize(vorgang, filterCriteria)) {
			throw new AccessDeniedException(String.format("Access denied to load vorgang %s", vorgangId));
		}

		return vorgang;
	}

	private boolean notAuthorize(Vorgang vorgang, FilterCriteria filterCriteria) {
		return !vorgangAuthenticationService.authorizeByOrganisationseinheitenId(vorgang, filterCriteria);
	}

	public Vorgang getById(String vorgangId) {
		var loaded = removeUnpermittedClientAttributes(loadById(vorgangId));

		return addLabels(loaded);
	}

	public Vorgang getOriginalVorgangById(String vorgangId) {
		return removeUnpermittedClientAttributes(loadById(vorgangId));
	}

	Vorgang removeUnpermittedClientAttributes(Vorgang vorgang) {
		return vorgang.toBuilder().clientAttributes(removeUnpermitted(vorgang.getClientAttributes())).build();
	}

	ClientAttributesMap removeUnpermitted(ClientAttributesMap attributes) {
		var result = new ClientAttributesMap();

		Optional.ofNullable(attributes).ifPresent(attri -> attri.entrySet().stream()
				.map(this::removeUnpermitted)
				.filter(pair -> !pair.getValue().isEmpty())
				.forEach(pair -> result.put(pair.getKey(), pair.getValue())));

		return result;
	}

	private Pair<String, ClientAttributeMap> removeUnpermitted(Entry<String, ClientAttributeMap> entry) {
		var resultAttributeMap = new ClientAttributeMap();
		entry.getValue().entrySet().stream()
				.filter(e -> readIsPermitted.test(e.getValue()))
				.forEach(e -> resultAttributeMap.put(e.getKey(), e.getValue()));

		return Pair.of(entry.getKey(), resultAttributeMap);
	}

	public void setAktenzeichen(Command command) {
		var aktenzeichen = StringUtils.trimToNull(MapUtils.getString(command.getBodyObject(), BODY_OBJECT_AKTENZEICHEN));
		repository.setAktenzeichen(command.getVorgangId(), command.getRelationVersion(), aktenzeichen);
	}

	public void assignToUser(Command command) {
		var patch = Map.<String, Object>of(Vorgang.MONGODB_FIELDNAME_ASSIGNED_TO, command.getBody().get(BODY_ASSIGNED_TO_FIELD));

		repository.patch(command.getRelationId(), command.getRelationVersion(), patch);

		publisher.publishEvent(new VorgangAssignedEvent(command));
	}

	public Stream<Vorgang> findAll() {
		return repository.findAll();
	}

	public boolean exists(String vorgangId, Collection<String> organisationEinheitenIds) {
		return repository.exists(vorgangId, organisationEinheitenIds);
	}

	Vorgang addLabels(Vorgang vorgang) {
		var mappedEingaenge = vorgang.getEingangs().stream().map(kopControlDataMapper::addLabels).toList();

		return vorgang.toBuilder().clearEingangs().eingangs(mappedEingaenge).build();
	}

	public void deleteVorgang(String vorgangId) {
		// TODO check version
		var stub = stubMapper.fromVorgang(loadById(vorgangId));

		repository.saveStub(stub);
	}

	Vorgang loadById(String vorgangId) {
		return repository.findById(vorgangId).orElseThrow(() -> new NotFoundException(Vorgang.class, vorgangId));
	}

	public void lockVorgang(Command command) {
		validateLockCommand(command);
		repository.patch(command.getVorgangId(), command.getRelationVersion(), buildLockPatch(command));
		publisher.publishEvent(new VorgangLockedEvent(command));
	}

	void validateLockCommand(Command command) {
		if (Objects.isNull(command.getCreatedByClientName())) {
			throw new TechnicalException("Missing client name in lock command");
		}
		if (Objects.isNull(getReason(command.getBodyObject()))) {
			throw new TechnicalException("Missing reason in lock command");
		}
	}

	Map<String, Object> buildLockPatch(Command command) {
		return Map.of(KEY_HEADER_LOCK, Map.of(
				Lock.FIELD_CLIENT_NAME, command.getCreatedByClientName(),
				Lock.FIELD_LOCKED_SINCE, ZonedDateTime.now(ZoneId.of("UTC")),
				Lock.FIELD_REASON, getReason(command.getBodyObject())));
	}

	String getReason(Map<String, Object> commandBody) {
		return StringUtils.trimToNull(MapUtils.getString(commandBody, Lock.FIELD_REASON));
	}

	public boolean isVorgangLocked(String vorgangId) {
		return repository.isVorgangLocked(vorgangId);
	}

	public void unlockVorgang(Command command) {
		validateUnlockingClient(command);
		repository.removeLock(command.getVorgangId(), command.getRelationVersion());
	}

	void validateUnlockingClient(Command command) {
		var vorgang = repository.findById(command.getVorgangId()).orElseThrow(() -> new NotFoundException(Vorgang.class, command.getVorgangId()));
		validateClientName(command.getCreatedByClientName(), Optional.ofNullable(vorgang.getHeader().getLock()));
	}

	void validateClientName(String commandClientName, Optional<Lock> lock) {
		if (hasNoPermissionToUnlock(commandClientName, lock)) {
			throw new AccessDeniedException(
					"Vorgang was locked by %s, but client name is %s!".formatted(lock.get().getClientName(), commandClientName));
		}
	}

	boolean hasNoPermissionToUnlock(String clientName, Optional<Lock> lock) {
		return lock.map(Lock::getClientName).filter(lockingClient -> !lockingClient.equals(clientName)).isPresent();
	}

}