/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.clientattribute;

import static org.assertj.core.api.Assertions.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.TestCallContextAttachingInterceptor;
import de.ozgcloud.vorgang.grpc.clientAttribute.ClientAttributeServiceGrpc.ClientAttributeServiceBlockingStub;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangRequestTestFactory;
import de.ozgcloud.vorgang.vorgang.GrpcFindVorgangWithEingangRequest;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangServiceGrpc.VorgangServiceBlockingStub;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(properties = {
		"grpc.server.inProcessName=test",
		"grpc.server.port=-1",
		"grpc.client.inProcess.address=in-process:test"
})
@DirtiesContext
@DataITCase
class ClientAttributeITCase {

	@GrpcClient("inProcess")
	private ClientAttributeServiceBlockingStub attributeServiceStub;
	@GrpcClient("inProcess")
	private VorgangServiceBlockingStub vorgangServiceStub;

	@Autowired
	private MongoTemplate mongoTemplate;

	private Vorgang vorgang;

	@BeforeEach
	void prepareDatabase() {
		mongoTemplate.dropCollection(Vorgang.class);
		vorgang = mongoTemplate.save(VorgangTestFactory.createBuilder().id(null).clientAttributes(null).build());
	}

	@BeforeEach
	void addInterceptor() {
		attributeServiceStub = attributeServiceStub.withInterceptors(new TestCallContextAttachingInterceptor());
		vorgangServiceStub = vorgangServiceStub.withInterceptors(new TestCallContextAttachingInterceptor());
	}

	@DisplayName("Test persisting ClientAttribute and load on Vorgang")
	@Nested
	class TestPersistAndLoad {

		@Test
		void shouldWorkWithoutAttributes() {
			var response = vorgangServiceStub.findVorgangWithEingang(GrpcFindVorgangWithEingangRequest.newBuilder().setId(vorgang.getId()).build());
			var vorgang = response.getVorgangWithEingang();

			assertThat(vorgang.getClientAttributesList()).isEmpty();
		}

		@Test
		void shouldLoadVorgangWithAttribute() {
			attributeServiceStub.set(GrpcSetClientAttributeRequestTestFactory.createBuilder().setVorgangId(vorgang.getId()).build());

			var response = vorgangServiceStub.findVorgangWithEingang(GrpcFindVorgangWithEingangRequest.newBuilder().setId(vorgang.getId()).build());
			var vorgang = response.getVorgangWithEingang();

			assertThat(vorgang.getClientAttributesList()).isNotEmpty();
		}

		@Test
		void shouldLoadVorgangListWithAttribute() {
			attributeServiceStub.set(GrpcSetClientAttributeRequestTestFactory.createBuilder().setVorgangId(vorgang.getId()).build());

			var response = vorgangServiceStub.findVorgang(GrpcFindVorgangRequestTestFactory.createFindAll());
			var vorgangList = response.getVorgangList();

			assertThat(vorgangList).hasSize(1);
			assertThat(vorgangList.get(0).getClientAttributesList()).describedAs("number of attributes").hasSize(1);
		}

	}

	@DisplayName("Test delete client attribute")
	@Nested
	class TestDelete {

		@BeforeEach
		void initDatabase() {
			mongoTemplate.save(VorgangTestFactory.create());
		}

		@Test
		void shouldDeleteInDatabase() {
			attributeServiceStub.delete(GrpcDeleteClientAttributeRequestTestFactory.create());

			var vorgangResponse = vorgangServiceStub
					.findVorgangWithEingang(GrpcFindVorgangWithEingangRequest.newBuilder().setId(VorgangTestFactory.ID).build());

			assertThat(vorgangResponse.getVorgangWithEingang().getClientAttributesList()).isEmpty();
		}
	}

}