/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.collaboration;

import static org.assertj.core.api.Assertions.*;
import static org.awaitility.Awaitility.*;
import static org.junit.jupiter.api.Assertions.fail;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

import org.bson.types.ObjectId;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.gridfs.GridFsTemplate;
import org.springframework.test.annotation.DirtiesContext;
import org.springframework.test.util.ReflectionTestUtils;

import com.google.protobuf.ByteString;
import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.apilib.client_attribute.grpc.OzgCloudClientAttributeMapper;
import de.ozgcloud.apilib.user.OzgCloudUserId;
import de.ozgcloud.apilib.user.OzgCloudUserProfile;
import de.ozgcloud.apilib.user.OzgCloudUserProfileService;
import de.ozgcloud.collaboration.CollaborationManagerConfiguration;
import de.ozgcloud.collaboration.CollaborationRequest;
import de.ozgcloud.collaboration.CollaborationService;
import de.ozgcloud.collaboration.CollaborationServiceGrpc.CollaborationServiceBlockingStub;
import de.ozgcloud.collaboration.GrpcGetFileContentRequest;
import de.ozgcloud.collaboration.request.CollaborationRequestId;
import de.ozgcloud.collaboration.request.CollaborationRequestServiceGrpc.CollaborationRequestServiceBlockingStub;
import de.ozgcloud.collaboration.request.GrpcFindRequestsRequest;
import de.ozgcloud.collaboration.request.GrpcGetRequestRequest;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandStatus;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.nachrichten.postfach.PostfachAddress;
import de.ozgcloud.nachrichten.postfach.PostfachNachricht;
import de.ozgcloud.nachrichten.postfach.PostfachRemoteService;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItem;
import de.ozgcloud.vorgang.attached_item.VorgangAttachedItemTestFactory;
import de.ozgcloud.vorgang.callcontext.CallContext;
import de.ozgcloud.vorgang.callcontext.TestCallContextAttachingInterceptor;
import de.ozgcloud.vorgang.callcontext.WithMockCustomUser;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.CreateCommandRequest;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.files.GridFsTestFactory;
import de.ozgcloud.vorgang.files.OzgFileTestFactory;
import de.ozgcloud.vorgang.vorgang.EingangTestFactory;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangHead;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import de.ozgcloud.vorgang.vorgang.ZustaendigeStelleTestFactory;
import io.grpc.Status;
import io.grpc.StatusRuntimeException;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(properties = {
		"grpc.server.inProcessName=test",
		"grpc.client.vorgang-manager.address=in-process:test",
		"grpc.client.ozgcloud-command-manager.address=in-process:test",
		"grpc.client.file-manager.address=in-process:test",
		"grpc.client.inProcess.address=in-process:test",
		"grpc.client.pluto.address=in-process:test"
})
@DataITCase
@WithMockCustomUser
@DirtiesContext(classMode = DirtiesContext.ClassMode.AFTER_EACH_TEST_METHOD)
class CollaborationITCase {

	private static final String FIELD_COLLABORATION_VORGANG_ID = "collaborationVorgangId";
	private static final String TITEL = "Collaboration Vorgang";
	private static final String BESCHREIBUNG = "Beschreibung der Anfrage";
	private static final String MUK_POSTFACH_ID = UUID.randomUUID().toString();

	@Autowired
	private CommandService commandService;
	@Autowired
	private CollaborationService collaborationService;

	@Autowired
	private MongoOperations mongoOperations;

	@MockBean
	@Qualifier(CollaborationManagerConfiguration.OZGCLOUD_USER_PROFILE_SERVICE_NAME)
	private OzgCloudUserProfileService collaborationOzgCloudUserProfileService;
	@MockBean
	private PostfachRemoteService postfachRemoteService;

	@Mock
	private OzgCloudUserProfile ozgCloudUserProfile;
	@MockBean
	private OzgCloudClientAttributeMapper clientAttributeMapper;

	private String vorgangId;

	@BeforeEach
	void init() {
		when(postfachRemoteService.getPostfachType()).thenReturn("MUK");
		mongoOperations.dropCollection(Command.class);
		mongoOperations.dropCollection(Vorgang.class);
		mongoOperations.dropCollection(VorgangAttachedItem.class);

		vorgangId = mongoOperations.save(VorgangTestFactory.createBuilder().id(null).version(0L).build()).getId();
	}

	@DisplayName("Get file content")
	@Nested
	class TestGetFileContent {

		@GrpcClient("inProcess")
		private CollaborationServiceBlockingStub collaborationStub;

		@Autowired
		private GridFsTemplate gridFsTemplate;

		@BeforeEach
		void initDatabase() {
			gridFsTemplate.delete(new Query());
		}

		@Test
		void shouldReturnStoredSmallFile() {
			var fileId = gridFsTemplate.store(GridFsTestFactory.createUpload());

			var file = downloadBinaryFile(fileId);

			assertThat(file).hasSameSizeAs(OzgFileTestFactory.CONTENT).isEqualTo(OzgFileTestFactory.CONTENT);
		}

		@Test
		void shouldReturnStoredFileWithSizeEqualChunkSize() {
			var content = OzgFileTestFactory.createContentInByte(FileService.CHUNK_SIZE);
			var fileId = gridFsTemplate.store(GridFsTestFactory.createUpload(content));

			var file = downloadBinaryFile(fileId);

			assertThat(file).hasSameSizeAs(content).isEqualTo(content);
		}

		@Test
		void shouldReturnStoredLargeFile() {
			var input = OzgFileTestFactory.createContentInMiB(5);
			var fileId = gridFsTemplate
					.store(GridFsTestFactory.createUpload(input));

			var downloadedFile = downloadBinaryFile(fileId);

			assertThat(downloadedFile).hasSameSizeAs(input).isEqualTo(input);
		}

		private byte[] downloadBinaryFile(ObjectId fileId) {
			var request = GrpcGetFileContentRequest.newBuilder()
					.setId(fileId.toHexString()).build();
			var it = getServiceStub().getFileContent(request);
			var content = new ByteArrayOutputStream();
			while (it.hasNext()) {
				ByteString chunkContent = it.next().getFileContent();
				try {
					content.write(chunkContent.toByteArray());
				} catch (IOException e) {
					fail(e);
				}
			}

			return content.toByteArray();
		}

		private CollaborationServiceBlockingStub getServiceStub() {
			return collaborationStub.withInterceptors(new TestCallContextAttachingInterceptor());
		}
	}

	private CreateCommandRequest buildCreateCollaborationVorgangCommand(String vorgangId, int collaborationLevel) {
		return CreateCommandRequest.builder()
				.callContext(CallContext.builder().client("test").build())
				.vorgangId(vorgangId)
				.relationId(vorgangId)
				.order("CREATE_COLLABORATION_REQUEST")
				.bodyObject(Map.of(
						"titel", TITEL,
						"beschreibung", BESCHREIBUNG,
						"collaborationLevel", collaborationLevel,
						"zustaendigeStelle", MUK_POSTFACH_ID))
				.build();
	}

	private List<VorgangAttachedItem> loadCollaborationRequest(String vorgangId) {
		var query = new Query(new Criteria().andOperator(
				Criteria.where(VorgangAttachedItem.FIELDNAME_VORGANG_ID).is(vorgangId),
				Criteria.where(VorgangAttachedItem.FIELDNAME_ITEM_NAME).is("CollaborationRequest")));
		return mongoOperations.find(query, VorgangAttachedItem.class);
	}

	private void waitUntilCommandHasStatus(String commandId, CommandStatus status) {
		await().atMost(600, TimeUnit.SECONDS).until(
				() -> mongoOperations.findById(commandId, Command.class),
				command -> command.getStatus() == status);
	}

	@Nested
	class TestCreateCollaborationVorgang {

		@BeforeEach
		void init() {
			when(ozgCloudUserProfile.getId()).thenReturn(OzgCloudUserId.from(CommandTestFactory.CREATED_BY));
			when(collaborationOzgCloudUserProfileService.getById(any())).thenReturn(ozgCloudUserProfile);
		}

		@Test
		void shouldSetCollaborationVorgangId() {
			var command = commandService.createCommand(
					buildCreateCollaborationVorgangCommand(vorgangId, CreateCollaborationVorgangRequestTestFactory.COLLABORATION_LEVEL));

			waitUntilCommandHasStatus(command.getId(), CommandStatus.FINISHED);

			var collaborationRequests = loadCollaborationRequest(vorgangId);
			assertThat(collaborationRequests).hasSize(1).first().extracting(VorgangAttachedItem::getItem, MAP)
					.containsKey(FIELD_COLLABORATION_VORGANG_ID);
		}

		@Test
		void shouldCreateCollaborationVorgang() {
			var command = commandService.createCommand(
					buildCreateCollaborationVorgangCommand(vorgangId, CreateCollaborationVorgangRequestTestFactory.COLLABORATION_LEVEL));

			waitUntilCommandHasStatus(command.getId(), CommandStatus.FINISHED);

			var collaborationVorgang = loadCollaborationVorgang(vorgangId);
			assertThat(collaborationVorgang.getHeader()).extracting(VorgangHead::getCollaborationLevel)
					.isEqualTo(CreateCollaborationVorgangRequestTestFactory.COLLABORATION_LEVEL);
			assertThat(collaborationVorgang.getHeader()).extracting(VorgangHead::getOrganisationsEinheitId)
					.isEqualTo(MUK_POSTFACH_ID);
			assertThat(collaborationVorgang.getClientAttributes()).isEmpty();
		}

		private Vorgang loadCollaborationVorgang(String sourceVorgangId) {
			var collaborationVorgangId = loadCollaborationRequest(sourceVorgangId).getFirst().getItem().get(FIELD_COLLABORATION_VORGANG_ID);
			return mongoOperations.findById(collaborationVorgangId, Vorgang.class);
		}

	}

	@Nested
	class TestCreateFachstellenBeteiligungRequest {

		private static final int COLLABORATION_LEVEL = 4;

		@Nested
		class TestSuccessfully {
			@Captor
			private ArgumentCaptor<PostfachNachricht> postfachNachrichtCaptor;

			@BeforeEach
			void init() {
				when(ozgCloudUserProfile.getId()).thenReturn(OzgCloudUserId.from(CommandTestFactory.CREATED_BY));
				when(collaborationOzgCloudUserProfileService.getById(any())).thenReturn(ozgCloudUserProfile);
			}

			@Test
			void shouldCreateCollaborationRequest() {
				var command = commandService.createCommand(buildCreateCollaborationVorgangCommand(vorgangId, COLLABORATION_LEVEL));

				waitUntilCommandHasStatus(command.getId(), CommandStatus.FINISHED);

				var collaborationRequests = loadCollaborationRequest(vorgangId);
				assertThat(collaborationRequests).hasSize(1).first().extracting(VorgangAttachedItem::getItem, MAP)
						.containsEntry(FIELD_COLLABORATION_VORGANG_ID, vorgangId);
			}

			@Test
			void shouldSendPostfachNachricht() {
				var command = commandService.createCommand(buildCreateCollaborationVorgangCommand(vorgangId, COLLABORATION_LEVEL));

				waitUntilCommandHasStatus(command.getId(), CommandStatus.FINISHED);

				verify(postfachRemoteService).sendMessage(postfachNachrichtCaptor.capture());
				assertThat(postfachNachrichtCaptor.getValue().getPostfachAddress()).extracting(PostfachAddress::getIdentifier)
						.hasToString(MUK_POSTFACH_ID);
			}

			@Test
			void shouldDeleteCollaborationRequest() {
				doThrow(TechnicalException.class).when(postfachRemoteService).sendMessage(any());
				var command = commandService.createCommand(buildCreateCollaborationVorgangCommand(vorgangId, COLLABORATION_LEVEL));

				waitUntilCommandHasStatus(command.getId(), CommandStatus.ERROR);

				assertThat(loadCollaborationRequest(vorgangId)).isEmpty();
			}
		}

		@Test
		void shouldFailOnMissingFachstelle() {
			ReflectionTestUtils.setField(collaborationService, "urlProvider", Optional.empty());
			var command = commandService.createCommand(buildCreateCollaborationVorgangCommand(vorgangId, COLLABORATION_LEVEL));

			waitUntilCommandHasStatus(command.getId(), CommandStatus.ERROR);

			command = mongoOperations.findById(command.getId(), Command.class);
			assertThat(command.getErrorMessage()).contains("Fachstelle not configured.");
		}
	}

	@DisplayName("Collaboration requests")
	@Nested
	class TestCollaborationRequests {

		@GrpcClient("inProcess")
		private CollaborationRequestServiceBlockingStub collaborationRequestStub;

		public static final String ID_STR = UUID.randomUUID().toString();
		public static final CollaborationRequestId ID = CollaborationRequestId.from(ID_STR);

		public static final String TITEL = LoremIpsum.getInstance().getWords(5);
		public static final String BESCHREIBUNG = LoremIpsum.getInstance().getWords(10);
		public static final String ZUSTAENDIGE_STELLE = UUID.randomUUID().toString();
		public static final String COLLABORATION_VORGANG_ID = UUID.randomUUID().toString();
		public static final int COLLABORATION_lEVEL = 1;

		private final Vorgang vorgang = VorgangTestFactory.create();

		private final Map<String, Object> collaborationRequestItem = Map.of(CollaborationRequest.PROPERTY_ID, ID_STR,
				CollaborationRequest.PROPERTY_COMMAND_ID, CommandTestFactory.ID,
				CollaborationRequest.PROPERTY_VORGANG_ID, VorgangTestFactory.ID,
				CollaborationRequest.PROPERTY_COLLABORATION_VORGANG_ID, COLLABORATION_VORGANG_ID,
				CollaborationRequest.PROPERTY_COLLABORATION_LEVEL, String.valueOf(COLLABORATION_lEVEL),
				CollaborationRequest.PROPERTY_CREATED_BY, CommandTestFactory.CREATED_BY,
				CollaborationRequest.PROPERTY_TITEL, TITEL,
				CollaborationRequest.PROPERTY_BESCHREIBUNG, BESCHREIBUNG,
				CollaborationRequest.PROPERTY_ZUSTAENDIGE_STELLE, ZUSTAENDIGE_STELLE);

		private final VorgangAttachedItem vorgangAttachedItem = VorgangAttachedItemTestFactory.createBuilder()
				.id(null)
				.vorgangId(VorgangTestFactory.ID)
				.version(0)
				.client("AlfaTestClient")
				.itemName("CollaborationRequest")
				.deleted(false)
				.item(collaborationRequestItem)
				.build();

		private VorgangAttachedItem savedCollaborationRequest;

		@BeforeEach
		void init() {
			mongoOperations.save(vorgang);
			savedCollaborationRequest = mongoOperations.save(vorgangAttachedItem);
		}

		@DisplayName("Get Request")
		@Nested
		class TestGetRequest {

			@Test
			void shouldReturnNotFoundExceptionOnNonExisting() {
				var id = UUID.randomUUID().toString();
				var request = GrpcGetRequestRequest.newBuilder().setId(id).build();
				var serviceStub = getServiceStub();

				assertThatThrownBy(() -> serviceStub.getRequest(request))
						.isInstanceOf(StatusRuntimeException.class)
						.hasMessageContaining("NOT_FOUND")
						.hasMessageContaining(id);
			}

			@Test
			void shouldReturnExistingCollaborationRequest() {
				var response = getServiceStub().getRequest(GrpcGetRequestRequest.newBuilder().setId(savedCollaborationRequest.getId()).build());

				assertThat(response.getRequest()).satisfies(grpcCollaborationRequest -> {
					assertThat(grpcCollaborationRequest.getId()).isEqualTo(savedCollaborationRequest.getId());
					assertThat(grpcCollaborationRequest.getTitel()).isEqualTo(TITEL);
					assertThat(grpcCollaborationRequest.getBeschreibung()).isEqualTo(BESCHREIBUNG);
					assertThat(grpcCollaborationRequest.getZustaendigeStelle()).isEqualTo(ZUSTAENDIGE_STELLE);
					assertThat(grpcCollaborationRequest.getCollaborationLevel()).isEqualTo(COLLABORATION_lEVEL);
					assertThat(grpcCollaborationRequest.getCreatedBy()).isEqualTo(CommandTestFactory.CREATED_BY);
				});
			}

			@DisplayName("Should throw access denied exception on non matching organisationsEinheitId")
			@Test
			void shouldThrowAccessDeniedException() {
				var vorgangId = mongoOperations.save(buildVorgang()).getId();
				savedCollaborationRequest = mongoOperations.save(vorgangAttachedItem.toBuilder().vorgangId(vorgangId).build());
				var request = GrpcGetRequestRequest.newBuilder().setId(savedCollaborationRequest.getId()).build();
				var serviceStub = getServiceStub();

				var exception = Assertions.assertThrows(StatusRuntimeException.class, () -> serviceStub.getRequest(request));
				assertThat(exception.getStatus().getCode()).isEqualTo(Status.Code.PERMISSION_DENIED);
			}

			private Vorgang buildVorgang() {
				return VorgangTestFactory.createBuilder().id(null).version(0).clearEingangs()
						.eingang(EingangTestFactory.createBuilder()
								.zustaendigeStelle(ZustaendigeStelleTestFactory.createBuilder().organisationseinheitenId("other").build())
								.build())
						.build();
			}
		}

		@DisplayName("Find requests")
		@Nested
		class TestFindRequests {

			@Test
			void shouldReturnExistingCollaborationRequests() {
				var response = getServiceStub().findRequests(GrpcFindRequestsRequest.newBuilder().setVorgangId(VorgangTestFactory.ID).build());

				assertThat(response.getRequestsList()).hasSize(1);
				assertThat(response.getRequestsList().getFirst().getId()).isEqualTo(savedCollaborationRequest.getId());
			}

			@Test
			void shouldReturnEmptyListOnNotMatchingVorgangId() {
				var response = getServiceStub().findRequests(GrpcFindRequestsRequest.newBuilder().setVorgangId(vorgangId).build());

				assertThat(response.getRequestsList()).isEmpty();
			}
		}

		private CollaborationRequestServiceBlockingStub getServiceStub() {
			return collaborationRequestStub.withInterceptors(new TestCallContextAttachingInterceptor());
		}
	}
}