/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.fail;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicBoolean;

import org.bson.types.ObjectId;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.data.mongodb.core.MongoOperations;
import org.springframework.data.mongodb.core.query.Criteria;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.data.mongodb.gridfs.GridFsResource;
import org.springframework.data.mongodb.gridfs.GridFsTemplate;
import org.springframework.test.annotation.DirtiesContext;

import com.google.protobuf.ByteString;
import com.mongodb.client.gridfs.model.GridFSFile;

import de.ozgcloud.common.test.DataITCase;
import de.ozgcloud.vorgang.callcontext.TestCallContextAttachingInterceptor;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceBlockingStub;
import de.ozgcloud.vorgang.grpc.binaryFile.BinaryFileServiceGrpc.BinaryFileServiceStub;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import de.ozgcloud.vorgang.vorgang.Vorgang;
import de.ozgcloud.vorgang.vorgang.VorgangTestFactory;
import io.grpc.StatusRuntimeException;
import io.grpc.stub.StreamObserver;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.RequiredArgsConstructor;
import net.devh.boot.grpc.client.inject.GrpcClient;

@SpringBootTest(properties = {
		"grpc.server.inProcessName=binary-file-test",
		"grpc.server.port=-1",
		"grpc.client.inProcess.address=in-process:binary-file-test"
})
@DirtiesContext
@DataITCase
public class BinaryFileITCase {

	@GrpcClient("inProcess")
	private BinaryFileServiceBlockingStub binaryFileServiceBlockingStub;
	@GrpcClient("inProcess")
	private BinaryFileServiceStub binaryFileServiceStub;

	@Autowired
	private GridFsTemplate gridFsTemplate;
	@Autowired
	private MongoOperations mongoOperations;

	@DisplayName("Find binary files meta data")
	@Nested
	class TestFindBinaryFilesMetaData {
		private ObjectId fileId;

		@BeforeEach
		void initDatabase() {
			gridFsTemplate.delete(new Query());
			fileId = gridFsTemplate.store(GridFsTestFactory.createUpload());
		}

		@Nested
		class OnMatchingOrganisationEinheitIdAtVorgang {

			@BeforeEach
			void initDatabase() {
				mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
				mongoOperations.save(VorgangTestFactory.create());
			}

			@Test
			void shouldReturnStoredFile() {
				var response = callFindBinaryFilesMetaData();

				assertThat(response.getFileList()).hasSize(1);

				var file = response.getFileList().get(0);

				assertThat(file.getId()).isEqualTo(fileId.toHexString());
				assertThat(file.getName()).isEqualTo(OzgFileTestFactory.NAME);
				assertThat(file.getContentType()).isEqualTo(OzgFileTestFactory.CONTENT_TYPE);
			}
		}

		@Nested
		class OnNonMatchingOrganisationEinheitIdOnVorgang {

			@BeforeEach
			void initDatabase() {
				mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("73"));
			}

			@Test
			void shouldThrowStatusRuntimeException() {
				assertThrows(StatusRuntimeException.class, () -> callFindBinaryFilesMetaData());
			}
		}

		private GrpcFindFilesResponse callFindBinaryFilesMetaData() {
			return prepareBinaryFileServiceBlockingStub().findBinaryFilesMetaData(
					GrpcBinaryFilesRequestTestFactory.createBuilder().clearFileId().addFileId(fileId.toHexString()).build());
		}

	}

	@Nested
	@DisplayName("Download files of different sizes")
	class TestDownloadFile {

		@BeforeEach
		void initDatabase() {
			gridFsTemplate.delete(new Query());
		}

		@Nested
		class OnMatchingOrganisationEinheitIdAtVorgang {

			@BeforeEach
			void initDatabase() {
				mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
				mongoOperations.save(VorgangTestFactory.create());
			}

			@Test
			void shouldReturnStoredSmallFile() {
				var fileId = gridFsTemplate.store(GridFsTestFactory.createUpload());

				var file = downloadBinaryFile(fileId);

				assertThat(file).hasSameSizeAs(OzgFileTestFactory.CONTENT).isEqualTo(OzgFileTestFactory.CONTENT);
			}

			@Test
			void shouldReturnStoredFileWithSizeEqualChunkSize() throws IOException {
				var content = OzgFileTestFactory.createContentInByte(FileService.CHUNK_SIZE);
				var fileId = gridFsTemplate.store(GridFsTestFactory.createUpload(content));

				var file = downloadBinaryFile(fileId);

				assertThat(file).hasSameSizeAs(content).isEqualTo(content);
			}

			@Test
			void shouldReturnStoredLargeFile() {
				var input = OzgFileTestFactory.createContentInMiB(5);
				var fileId = gridFsTemplate
						.store(GridFsTestFactory.createUpload(input));

				var downloadedFile = downloadBinaryFile(fileId);

				assertThat(downloadedFile).hasSameSizeAs(input).isEqualTo(input);
			}
		}

		@Nested
		class OnNonMatchingOrganisationEinheitIdAtVorgang {

			@BeforeEach
			void initDatabase() {
				mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("73"));
			}

			@Test
			void shouldThrowStatusRuntimeException() {
				var fileId = gridFsTemplate.store(GridFsTestFactory.createUpload());

				assertThrows(StatusRuntimeException.class, () -> downloadBinaryFile(fileId));
			}
		}

		private byte[] downloadBinaryFile(ObjectId fileId) {
			GrpcGetBinaryFileDataRequest request = GrpcGetBinaryFileDataRequest.newBuilder()
					.setFileId(fileId.toHexString()).build();
			Iterator<GrpcGetBinaryFileDataResponse> it = prepareBinaryFileServiceBlockingStub().getBinaryFileContent(request);
			var content = new ByteArrayOutputStream();
			while (it.hasNext()) {
				ByteString chunkContent = it.next().getFileContent();
				try {
					content.write(chunkContent.toByteArray());
				} catch (IOException e) {
					fail(e);
				}
			}

			return content.toByteArray();
		}
	}

	private BinaryFileServiceBlockingStub prepareBinaryFileServiceBlockingStub() {
		return binaryFileServiceBlockingStub.withInterceptors(new TestCallContextAttachingInterceptor());
	}

	@Nested
	@DisplayName("Upload file")
	class TestFileUpload {
		private int onNextCount = 0;
		private AtomicBoolean hasError = new AtomicBoolean(false);
		private AtomicBoolean completed = new AtomicBoolean(false);
		private GrpcUploadBinaryFileMetaData metaData = GrpcUploadBinaryFileMetaDataTestFactory.create();
		private ByteString fileContent = ByteString.copyFrom(OzgFileTestFactory.CONTENT);
		private StubStreamObserver responseObserver;

		@RequiredArgsConstructor(access = AccessLevel.PROTECTED)
		class StubStreamObserver implements StreamObserver<GrpcUploadBinaryFileResponse> {
			private final CompletableFuture<FileId> fileIdFuture;
			@Getter
			private String fileId;

			@Override
			public void onNext(GrpcUploadBinaryFileResponse value) {
				fileId = value.getFileId();
				onNextCount += 1;
			}

			@Override
			public void onError(Throwable t) {
				hasError.set(true);
				fileIdFuture.completeExceptionally(t);
			}

			@Override
			public void onCompleted() {
				completed.set(true);
				fileIdFuture.complete(FileId.from(fileId));
			}
		};

		@BeforeEach
		void initDatabase() {
			gridFsTemplate.delete(new Query());
		}

		@Nested
		class OnMatchingOrganisationEinheitIdAtVorgang {
			@BeforeEach
			void init() {
				mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
				mongoOperations.save(VorgangTestFactory.create());
			}

			@Test
			void shouldCallOnNext() {
				var fileIdFuture = saveFileWithContent();

				assertThat(fileIdFuture).isCompleted();
				assertThat(onNextCount).isEqualTo(1);
			}

			@Test
			void shouldCallOnComplete() {
				var fileIdFuture = saveFileWithContent();

				assertThat(fileIdFuture).isCompleted();
				assertThat(completed).isTrue();
			}

			@Test
			void shouldHaveFileId() {
				var fileIdFuture = saveFileWithContent();

				assertThat(fileIdFuture).isCompleted().isDone().isNotNull();
			}

			@Test
			void shouldSaveVorgangId() {
				var fileIdFuture = saveFileWithContent();

				assertThat(fileIdFuture).isCompleted();
				GridFSFile file = gridFsTemplate.findOne(Query.query(Criteria.where("metadata.vorgangId").is(VorgangTestFactory.ID)));
				assertThat(file.getMetadata()).containsEntry("vorgangId", VorgangTestFactory.ID);
			}

			@Test
			void shouldSaveFileContent() throws Exception {
				var fileIdFuture = saveFileWithContent();

				assertThat(fileIdFuture).isCompleted();
				GridFSFile file = gridFsTemplate.findOne(Query.query(Criteria.where("metadata.vorgangId").is(VorgangTestFactory.ID)));
				GridFsResource resource = gridFsTemplate.getResource(file);
				try (InputStream fileStream = resource.getContent()) {
					byte[] fileContentFromDb = fileStream.readAllBytes();
					assertThat(fileContentFromDb).hasSameSizeAs(fileContent.toByteArray()).isEqualTo(fileContent.toByteArray());
				}

			}
		}

		@Nested
		class OnMatchingNonOrganisationEinheitIdAtVorgang {
			@BeforeEach
			void init() {
				mongoOperations.dropCollection(Vorgang.COLLECTION_NAME);
				mongoOperations.save(VorgangTestFactory.createWithOrganisationEinheitId("73"));
			}

			@Test
			void shouldThowStatusRuntimeException() {
				var fileIdFuture = saveFileWithContent();
				assertThat(fileIdFuture).isCompletedExceptionally().isNotCancelled().failsWithin(1, TimeUnit.SECONDS)
						.withThrowableOfType(ExecutionException.class).withCauseInstanceOf(StatusRuntimeException.class);
			}
		}

		private CompletableFuture<FileId> saveFileWithContent() {
			var fileIdFuture = new CompletableFuture<FileId>();
			GrpcUploadBinaryFileRequest request1 = GrpcUploadBinaryFileRequest.newBuilder().setMetadata(metaData).build();
			GrpcUploadBinaryFileRequest request2 = GrpcUploadBinaryFileRequest.newBuilder().setFileContent(fileContent).build();

			responseObserver = new StubStreamObserver(fileIdFuture);

			StreamObserver<GrpcUploadBinaryFileRequest> observer = prepareBinaryFileServiceStub().uploadBinaryFileAsStream(responseObserver);

			observer.onNext(request1);
			observer.onNext(request2);
			observer.onCompleted();

			try {
				fileIdFuture.get(10, TimeUnit.SECONDS);
			} catch (InterruptedException | ExecutionException | TimeoutException e) {
				// ignored, the right exception will be checked in the test case
			}

			return fileIdFuture;

		}

		private BinaryFileServiceStub prepareBinaryFileServiceStub() {
			return binaryFileServiceStub.withInterceptors(new TestCallContextAttachingInterceptor());
		}
	}
}