/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.files;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Collection;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicBoolean;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.ArgumentMatchers;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.vorgang.common.security.PolicyService;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcFindFilesResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcGetBinaryFileDataResponse;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileMetaData;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileRequest;
import de.ozgcloud.vorgang.grpc.binaryFile.GrpcUploadBinaryFileResponse;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.ServerCallStreamObserver;
import io.grpc.stub.StreamObserver;

class GrpcBinaryFileServiceTest {

	@InjectMocks
	@Spy
	private GrpcBinaryFileService grpcService;
	@Mock
	private FileService service;
	@Mock
	private GrpcOzgFileMapper fileMapper;
	@Mock
	private PolicyService policyService;

	@Nested
	class TestFindBinaryFilesMetaData {

		@Mock
		private StreamObserver<GrpcFindFilesResponse> responseObserver;
		@Captor
		private ArgumentCaptor<Collection<FileId>> fileIdsCaptor;
		@Captor
		private ArgumentCaptor<GrpcFindFilesResponse> responseCaptor;

		@Test
		void shouldCallPolicyService() {
			findBinaryFilesMetaData();

			verify(policyService).checkPermission(ArgumentMatchers.<Set<String>>any());
		}

		@Test
		void shouldCallService() {
			findBinaryFilesMetaData();

			verify(service).findFilesMetaData(fileIdsCaptor.capture());
			assertThat(fileIdsCaptor.getValue()).contains(OzgFileTestFactory.ID);
		}

		@Test
		void shouldReturnList() {
			findBinaryFilesMetaData();

			verify(responseObserver).onNext(responseCaptor.capture());
		}

		@Test
		void shouldCallComplete() {
			findBinaryFilesMetaData();

			verify(responseObserver).onCompleted();
		}

		private void findBinaryFilesMetaData() {
			grpcService.findBinaryFilesMetaData(GrpcBinaryFilesRequestTestFactory.create(), responseObserver);

			verify(responseObserver).onNext(responseCaptor.capture());
		}
	}

	@DisplayName("Upload filestream async")
	@Nested
	class TestUploadFileStreamAsync {

		private static final int FILE_SIZE = 5 * 1_000_000;

		@Mock
		private StreamObserver<GrpcFindFilesResponse> responseObserver;

		@Spy
		private ServerCallStreamObserver<GrpcUploadBinaryFileResponse> uploadFileResponseObserver;

		@Mock
		private StreamObserver<GrpcGetBinaryFileDataResponse> dataResponseObserver;

		@BeforeEach
		void init() {
			when(service.uploadFileStreamAsync(any(), any(), any(), any())).thenReturn(CompletableFuture.supplyAsync(() -> FileId.createNew()));
		}

		@Test
		void shouldCallService() throws IOException {
			uploadFile(FILE_SIZE);

			verify(service, timeout(500)).uploadFileStreamAsync(any(), any(), any(), any());
		}

		@Test
		void shouldCallOnComplete() throws IOException {
			uploadFile(FILE_SIZE);

			verify(uploadFileResponseObserver, timeout(500)).onCompleted();
		}

		@Test
		void shouldCallOnNext() throws IOException {
			uploadFile(FILE_SIZE);

			verify(uploadFileResponseObserver, timeout(500)).onNext(any());
		}

		void uploadFile(int fileSize) {
			var observer = grpcService.uploadBinaryFileAsStream(uploadFileResponseObserver);

			try {
				uploadChunks(fileSize, observer).get();
			} catch (InterruptedException | ExecutionException e) {
				Thread.currentThread().interrupt();
			}
		}

		private CompletableFuture<Boolean> uploadChunks(final int fileSize, final StreamObserver<GrpcUploadBinaryFileRequest> observer) {
			return CompletableFuture.supplyAsync(() -> {
				GrpcUploadBinaryFileRequest firstRequest = GrpcUploadBinaryFileRequest.newBuilder()
						.setMetadata(GrpcUploadBinaryFileMetaData.newBuilder()
								.setFileName("test")
								.build())
						.build();

				observer.onNext(firstRequest);

				observer.onCompleted();

				return Boolean.TRUE;
			});
		}
	}

	@Nested
	class TestGetFileStream {
		private static final int FILE_SIZE_MB = 5;

		@Mock
		private StreamObserver<GrpcFindFilesResponse> responseObserver;

		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> dataResponseObserver;

		private GrpcGetBinaryFileDataRequest request;

		@BeforeEach
		void init() {
			request = GrpcGetBinaryFileDataRequest.newBuilder().setFileId(OzgFileTestFactory.ID.toString()).build();
		}

		@Test
		void shouldCallService() {
			when(service.getUploadedFileStream(any(FileId.class)))
					.thenReturn(new ByteArrayInputStream(OzgFileTestFactory.createContentInMiB(FILE_SIZE_MB)));

			grpcService.getBinaryFileContent(request, dataResponseObserver);

			verify(service, timeout(500)).getUploadedFileStream(any(FileId.class));
		}

		@Test
		void shouldCallSetOnReadyHander() {
			grpcService.getBinaryFileContent(request, dataResponseObserver);

			verify(dataResponseObserver, timeout(500)).setOnReadyHandler(any());
		}
	}

	@Nested
	class TestGetBinaryFileData {

		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;
		@Captor
		private ArgumentCaptor<Collection<FileId>> fileIdsCaptor;
		@Captor
		private ArgumentCaptor<GrpcGetBinaryFileDataResponse> responseCaptor;

		@BeforeEach
		void mock() {
			when(service.getUploadedFileStream(any(FileId.class))).thenReturn(new ByteArrayInputStream(OzgFileTestFactory.CONTENT));
		}

		@Test
		void shouldCallPolicyService() {
			getBinaryFileData();

			verify(policyService).checkPermissionByFile(anyString());
		}

		@Test
		void shouldCallService() {
			getBinaryFileData();

			verify(service).getUploadedFileStream(FileId.from(GrpcGetBinaryFileDataRequestTestFactory.FILE_ID));
		}

		@Test
		void shouldCallSetOnReadyHandler() {
			getBinaryFileData();

			verify(responseObserver).setOnReadyHandler(any());
		}

		@Test
		void shouldCallOnNext() {
			getBinaryFileData();

			verify(responseObserver, times(1)).onNext(any());
		}

		@Test
		void shouldCallOnComplete() {
			getBinaryFileData();

			verify(responseObserver).onCompleted();
		}

		private void getBinaryFileData() {
			AtomicBoolean doSendFile = new AtomicBoolean(true);
			grpcService.getBinaryFileContent(GrpcGetBinaryFileDataRequestTestFactory.create(), responseObserver);

			grpcService.sendChunks(responseObserver, new ByteArrayInputStream(OzgFileTestFactory.CONTENT), doSendFile);

		}
	}

	@Nested
	class TestBuildChunkResponse {
		@Test
		void shouldHaveFileContent() {
			GrpcGetBinaryFileDataResponse response = grpcService.buildChunkResponse(OzgFileTestFactory.CONTENT,
					OzgFileTestFactory.CONTENT.length);

			assertThat(response.getFileContent()).isNotEmpty();
		}
	}

	@Nested
	class TestBuildUploadResponse {
		@Test
		void shouldHaveFileId() {
			GrpcUploadBinaryFileResponse response = GrpcBinaryFileService.buildResponse(FileId.createNew());

			assertThat(response.getFileId()).isNotNull();
		}
	}

	@Nested
	class TestDownloadFile {
		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;

		@Test
		void shouldSetOnReadyHandler() {
			grpcService.downloadFile(responseObserver, new ByteArrayInputStream(OzgFileTestFactory.createContentInMiB(3)));

			verify(responseObserver).setOnReadyHandler(any());
		}
	}

	@Nested
	class TestOnReadyHandler {
		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;

		@Test
		void shouldSetOnReadyHandler() {
			grpcService.downloadFile(responseObserver, new ByteArrayInputStream(OzgFileTestFactory.createContentInMiB(3)));

			verify(responseObserver).setOnReadyHandler(any());
		}
	}

	@Nested
	class TestSendChunk {
		@Mock
		private CallStreamObserver<GrpcGetBinaryFileDataResponse> responseObserver;

		@Test
		void shouldComplete() throws IOException {
			var res = OzgFileTestFactory.createContentInMiB(1);
			var inputStream = new ByteArrayInputStream(res);

			grpcService.downloadFile(responseObserver, inputStream);

			triggerSendingChunks(res, inputStream);

			verify(responseObserver, timeout(500)).onCompleted();
		}

		@Test
		void shoudSendChunks() {
			grpcService.sendChunks(responseObserver, new ByteArrayInputStream(OzgFileTestFactory.createContentInMiB(3)), new AtomicBoolean(true));

			verify(grpcService).sendChunks(any(), any(), any());
		}

		@Test
		void shouldSendOneChunk() {
			grpcService.sendChunks(responseObserver, new ByteArrayInputStream(OzgFileTestFactory.createContentInMiB(3)), new AtomicBoolean(true));

			verify(responseObserver).onNext(any());
		}

		@Test
		void shouldSendOneChunkWhenFileSizeEqualsChunkSize() throws IOException {
			byte[] res = OzgFileTestFactory.createContentInByte(FileService.CHUNK_SIZE);
			var inputStream = new ByteArrayInputStream(res);

			grpcService.downloadFile(responseObserver, inputStream);

			triggerSendingChunks(res, inputStream);

			verify(grpcService).buildChunkResponse(any(), anyInt());
			verify(responseObserver).onNext(any());
			verify(responseObserver, timeout(500)).onCompleted();
		}

		@Test
		void shouldSendManyChunks() throws IOException {
			byte[] res = OzgFileTestFactory.createContentInMiB(1);
			var inputStream = new ByteArrayInputStream(res);

			grpcService.downloadFile(responseObserver, inputStream);

			int expected = triggerSendingChunks(res, inputStream);

			verify(grpcService, times(expected)).buildChunkResponse(any(), anyInt());
			verify(responseObserver, times(expected)).onNext(any());
		}

		private int triggerSendingChunks(byte[] res, ByteArrayInputStream inputStream) {
			int expected = computeExpectedCunkCount(res);
			AtomicBoolean doSendFile = new AtomicBoolean(true);
			for (int i = 0; i < expected; i++) {
				grpcService.sendChunks(responseObserver, inputStream, doSendFile);
			}
			if (res.length % FileService.CHUNK_SIZE == 0) {
				grpcService.sendChunks(responseObserver, inputStream, doSendFile);
			}
			return expected;
		}

		int computeExpectedCunkCount(byte[] res) {
			int expected = res.length / FileService.CHUNK_SIZE;
			if (res.length % FileService.CHUNK_SIZE != 0) {
				expected++;
			}
			return expected;
		}
	}
}