/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.vorgang.vorgang;

import static de.ozgcloud.vorgang.vorgang.VorgangService.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Collections;
import java.util.Map;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.springframework.context.ApplicationEventPublisher;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEvent;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandRevokeFailedEvent;
import de.ozgcloud.command.CommandRevokedEvent;
import de.ozgcloud.command.RevokeCommandEvent;
import de.ozgcloud.command.VorgangUnlockedEvent;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.vorgang.command.CommandCreatedEventTestFactory;
import de.ozgcloud.vorgang.command.CommandService;
import de.ozgcloud.vorgang.command.CommandTestFactory;
import de.ozgcloud.vorgang.command.Order;
import de.ozgcloud.vorgang.common.errorhandling.FunctionalException;
import de.ozgcloud.vorgang.files.FileService;
import de.ozgcloud.vorgang.status.StatusService;
import de.ozgcloud.vorgang.vorgang.redirect.VorgangForwardFailedEventTestFactory;
import de.ozgcloud.vorgang.vorgang.redirect.VorgangRedirectedEventTestFactory;

class VorgangEventListenerTest {

	@InjectMocks
	private VorgangEventListener listener;

	@Mock
	private VorgangService service;
	@Mock
	private VorgangHeaderService vorgangHeaderService;
	@Mock
	private StatusService statusService;
	@Mock
	private FileService fileService;
	@Mock
	private CommandService commandService;

	@Mock
	private ApplicationEventPublisher publisher;

	@Nested
	class OnVorgangRedirectedEvent {

		@Test
		void shouldSetStatus() {
			listener.updateStatus(VorgangRedirectedEventTestFactory.create());

			verify(statusService).setStatusToWeitergeleitet(VorgangRedirectedEventTestFactory.SOURCE.getVorgangId(),
					VorgangRedirectedEventTestFactory.SOURCE.getCreatedByCommand());
		}
	}

	@Nested
	class TestUpdateStatusOnVorgangForwardFailedEvent {

		@Test
		void shouldSetStatus() {
			listener.updateStatus(VorgangForwardFailedEventTestFactory.create());

			verify(statusService).setStatusToInBearbeitung(VorgangForwardFailedEventTestFactory.SOURCE.getVorgangId(),
					VorgangForwardFailedEventTestFactory.SOURCE.getId());
		}
	}

	@Nested
	class TestDeleteVorgang {

		@Test
		void shouldDeleteVorgang() {
			listener.onVorgangLoeschen(CommandCreatedEventTestFactory.create());

			verify(service).deleteVorgang(VorgangTestFactory.ID);
		}

		@Test
		void shouldPublishDoneEvent() {
			listener.onVorgangLoeschen(CommandCreatedEventTestFactory.create());

			verify(publisher).publishEvent(any(CommandExecutedEvent.class));
		}

		@Test
		void shouldPublishFailedEvent() {
			doThrow(new FunctionalException(() -> "test")).when(service).deleteVorgang(any());

			listener.onVorgangLoeschen(CommandCreatedEventTestFactory.create());

			verify(publisher).publishEvent(any(CommandFailedEvent.class));
		}

		@Test
		void shouldDeleteFiles() {
			listener.onVorgangLoeschen(CommandCreatedEventTestFactory.create());

			verify(fileService).deleteAllByVorgang(VorgangTestFactory.ID);
		}
	}

	@Nested
	class TestSetAktenzeichen {

		private static final Command SET_AKTENZEICHEN_COMMAND = CommandTestFactory.createBuilder().order(Order.SET_AKTENZEICHEN.name())
				.bodyObject(Map.of(BODY_OBJECT_AKTENZEICHEN, VorgangTestFactory.AKTENZEICHEN)).build();
		private static final String PREVIOUS_AKTENZEICHEN = VorgangTestFactory.ID + "-prev";

		@BeforeEach
		void init() {
			when(vorgangHeaderService.getById(VorgangTestFactory.ID)).thenReturn(
					VorgangHeaderTestFactory.createBuilder().aktenzeichen(PREVIOUS_AKTENZEICHEN).build());
		}

		@Test
		void shouldCallGetById() {
			callListener();

			verify(vorgangHeaderService).getById(VorgangTestFactory.ID);
		}

		@Test
		void shouldFailIfVorgangNotFound() {
			when(vorgangHeaderService.getById(VorgangTestFactory.ID)).thenThrow(TechnicalException.class);

			callListener();

			verify(publisher).publishEvent(any(CommandFailedEvent.class));
		}

		@Test
		void shouldSetPreviousAktenzeichen() {
			callListener();

			verify(commandService).setPreviousState(CommandTestFactory.ID,
					Map.of(VorgangService.BODY_OBJECT_AKTENZEICHEN, PREVIOUS_AKTENZEICHEN));
		}

		@Test
		@DisplayName("should store a null in previousState, if Aktenzeichen is null")
		void shouldHandleNullAktenzeichen() {
			when(vorgangHeaderService.getById(VorgangTestFactory.ID)).thenReturn(VorgangHeaderTestFactory.createBuilder().aktenzeichen(null).build());

			callListener();

			verify(commandService).setPreviousState(CommandTestFactory.ID, Collections.singletonMap(VorgangService.BODY_OBJECT_AKTENZEICHEN, null));
		}

		@Test
		void shouldCallVorgangService() {
			callListener();

			verify(service).setAktenzeichen(SET_AKTENZEICHEN_COMMAND);
		}

		@Test
		void shouldPublishEvent() {
			callListener();

			verify(publisher).publishEvent(any(SetAktenzeichenCompletedEvent.class));
		}

		@Nested
		class TestFailedEvent {

			private final String ERROR_MESSAGE = LoremIpsum.getInstance().getWords(3);

			@Captor
			private ArgumentCaptor<CommandFailedEvent> event;

			@BeforeEach
			void init() {
				doThrow(new RuntimeException(ERROR_MESSAGE)).when(service).setAktenzeichen(any());
			}

			@Test
			void shouldContainCommandId() {
				callListener();

				verify(publisher).publishEvent(event.capture());
				assertThat(event.getValue().getSource()).isEqualTo(CommandTestFactory.ID);
				assertThat(event.getValue().getErrorMessage()).isEqualTo(ERROR_MESSAGE);
			}

			@Test
			void shouldContainMessage() {
				callListener();

				verify(publisher).publishEvent(event.capture());
				assertThat(event.getValue().getErrorMessage()).isEqualTo(ERROR_MESSAGE);
			}
		}

		private void callListener() {
			listener.onSetAktenzeichen(CommandCreatedEventTestFactory.create(SET_AKTENZEICHEN_COMMAND));
		}
	}

	@Nested
	class TestOnLockVorgang {

		@Captor
		private ArgumentCaptor<CommandFailedEvent> eventCaptor;

		@Test
		void shouldCallLockVorgang() {
			var command = CommandTestFactory.create();

			listener.onLockVorgang(CommandCreatedEventTestFactory.create(command));

			verify(service).lockVorgang(command);
		}

		@Test
		void shouldPublishCommandFailedEvent() {
			var command = CommandTestFactory.create();
			var errorMessage = "error message";
			doThrow(new RuntimeException(errorMessage)).when(service).lockVorgang(command);

			listener.onLockVorgang(CommandCreatedEventTestFactory.create(command));

			verify(publisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue()).satisfies(event -> {
				assertThat(event.getSource()).isEqualTo(CommandTestFactory.ID);
				assertThat(event.getErrorMessage()).isEqualTo(errorMessage);
			});
		}
	}

	@Nested
	class TestRevokeLockVorgang {

		@Captor
		private ArgumentCaptor<CommandRevokedEvent> commandRevokedEventCaptor;
		@Captor
		private ArgumentCaptor<CommandRevokeFailedEvent> commandRevokeFailedEventCaptor;

		@Test
		void shouldCallUnlockVorgang() {
			var command = CommandTestFactory.create();

			listener.onRevokeLockVorgang(new RevokeCommandEvent(command));

			verify(service).unlockVorgang(command);
		}

		@Test
		void shouldPublishCommandRevokedEvent() {
			var command = CommandTestFactory.create();

			listener.onRevokeLockVorgang(new RevokeCommandEvent(command));

			verify(publisher).publishEvent(commandRevokedEventCaptor.capture());
			assertThat(commandRevokedEventCaptor.getValue().getSource()).isSameAs(command);
		}

		@Test
		void shouldPublishCommandFailedEvent() {
			var command = CommandTestFactory.create();
			var errorMessage = "error message";
			doThrow(new RuntimeException(errorMessage)).when(service).unlockVorgang(command);

			listener.onRevokeLockVorgang(new RevokeCommandEvent(command));

			verify(publisher).publishEvent(commandRevokeFailedEventCaptor.capture());
			assertThat(commandRevokeFailedEventCaptor.getValue()).satisfies(event -> {
				assertThat(event.getSource()).isEqualTo(CommandTestFactory.ID);
				assertThat(event.getErrorMessage()).isEqualTo(errorMessage);
			});
		}
	}

	@Nested
	class TestOnUnlockVorgang {

		private final Command command = CommandTestFactory.create();

		@Captor
		private ArgumentCaptor<VorgangUnlockedEvent> vorgangUnlockedEventCaptor;
		@Captor
		private ArgumentCaptor<CommandFailedEvent> commandFailedEventCaptor;

		@Test
		void shouldUnlockVorgangOnUnlockVorgangOrder() {
			onUnlockVorgang();

			verify(service).unlockVorgang(command);
		}

		@Test
		void shouldPublishVorgangUnlockedEvent() {
			onUnlockVorgang();

			verify(publisher).publishEvent(vorgangUnlockedEventCaptor.capture());
			assertThat(vorgangUnlockedEventCaptor.getValue()).satisfies(event -> {
				assertThat(event).isInstanceOf(VorgangUnlockedEvent.class);
				assertThat(event.getCommand()).isEqualTo(command);
			});

		}

		@Test
		void shouldPublishCommandFailedEvent() {
			var errorMessage = "error message";
			doThrow(new RuntimeException(errorMessage)).when(service).unlockVorgang(command);

			onUnlockVorgang();

			verify(publisher).publishEvent(commandFailedEventCaptor.capture());
			assertThat(commandFailedEventCaptor.getValue()).satisfies(event -> {
				assertThat(event.getSource()).isEqualTo(CommandTestFactory.ID);
				assertThat(event.getErrorMessage()).isEqualTo(errorMessage);
			});
		}

		private void onUnlockVorgang() {
			listener.onUnlockVorgang(new CommandCreatedEvent(command));
		}
	}
}