package de.ozgcloud.vorgang.common.grpc;

import static de.ozgcloud.vorgang.common.grpc.FormDataMappingUtils.*;
import static de.ozgcloud.vorgang.common.grpc.GrpcSubFormTestFactory.*;
import static org.assertj.core.api.Assertions.*;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;

import de.ozgcloud.vorgang.vorgang.GrpcFormField;

class FormDataMappingUtilsTest {

	@Nested
	class TestCombiner {

		@Test
		void missingInMap2ShouldContainValue() {
			Map<String, Object> map1 = Map.of("TEST", "VALUE1");
			Map<String, Object> map2 = Map.of();

			var combined = FormDataMappingUtils.combiner(map1, map2);

			assertThat(combined).containsEntry("TEST", "VALUE1");
		}

		@Test
		void missingInMap1ShouldContainValue() {
			Map<String, Object> map1 = new HashMap<>();
			Map<String, Object> map2 = Map.of("TEST", "VALUE1");

			var combined = FormDataMappingUtils.combiner(map1, map2);

			assertThat(combined).containsEntry("TEST", "VALUE1");
		}

		@Test
		void inBothMapShouldCombineToList() {
			Map<String, Object> map1 = new HashMap<>(Map.of("TEST", "VALUE1"));
			Map<String, Object> map2 = Map.of("TEST", "VALUE2");

			var combined = FormDataMappingUtils.combiner(map1, map2);

			assertThat(combined).containsEntry("TEST", List.of("VALUE1", "VALUE2"));
		}

		@Test
		void listInMap2ShouldStayAList() {
			Map<String, Object> map1 = Map.of("TEST", List.of("VALUE1", "VALUE2"));
			Map<String, Object> map2 = Map.of();

			var combined = FormDataMappingUtils.combiner(map1, map2);

			assertThat(combined).containsEntry("TEST", List.of("VALUE1", "VALUE2"));
		}

		@Test
		void valueInMap1ShouldBeAddedToList() {
			Map<String, Object> map1 = new HashMap<>(Map.of("TEST", "VALUE1"));
			Map<String, Object> map2 = Map.of("TEST", List.of("VALUE2", "VALUE3"));

			var combined = FormDataMappingUtils.combiner(map1, map2);

			assertThat(combined).containsEntry("TEST", List.of("VALUE1", "VALUE2", "VALUE3"));
		}

		@Test
		void valueInMap2ShouldBeAddedToList() {
			Map<String, Object> map1 = new HashMap<>(Map.of("TEST", List.of("VALUE1", "VALUE2")));
			Map<String, Object> map2 = Map.of("TEST", "VALUE3");

			var combined = FormDataMappingUtils.combiner(map1, map2);

			assertThat(combined).containsEntry("TEST", List.of("VALUE1", "VALUE2", "VALUE3"));
		}

		@Test
		void listInMap1ShouldStayAList() {
			Map<String, Object> map1 = new HashMap<>();
			Map<String, Object> map2 = Map.of("TEST", List.of("VALUE1", "VALUE2"));

			var combined = FormDataMappingUtils.combiner(map1, map2);

			assertThat(combined).containsEntry("TEST", List.of("VALUE1", "VALUE2"));
		}

		@Test
		void listInBothMapsShouldByCombined() {
			Map<String, Object> map1 = new HashMap<>(Map.of("TEST", List.of("VALUE1", "VALUE2")));
			Map<String, Object> map2 = Map.of("TEST", List.of("VALUE3", "VALUE4"));

			var combined = FormDataMappingUtils.combiner(map1, map2);

			assertThat(combined).containsEntry("TEST", List.of("VALUE1", "VALUE2", "VALUE3", "VALUE4"));
		}
	}

	@Nested
	@DisplayName("Add value to map")
	class TestAddValueToMap {

		private static final String KEY = "key";

		private Map<String, Object> map = new HashMap<>();

		@Test
		void shouldAddValue() {
			FormDataMappingUtils.addToMap(map, KEY, "value");

			assertThat(map).containsEntry("key", "value");
		}

		@Test
		void shouldAddCollection() {
			FormDataMappingUtils.addToMap(map, KEY, List.of("value1", "value2"));

			assertThat(map).containsEntry("key", List.of("value1", "value2"));
		}

		@Test
		void shouldAddCollection2() {
			map.put(KEY, "value");

			FormDataMappingUtils.addToMap(map, KEY, List.of("value1", "value2"));

			assertThat(map).containsEntry("key", List.of("value", "value1", "value2"));
		}

		@Test
		void shouldAddCollection3() {
			map.put(KEY, List.of("value"));

			FormDataMappingUtils.addToMap(map, KEY, List.of("value1", "value2"));

			assertThat(map).containsEntry("key", List.of("value", "value1", "value2"));
		}

		@Test
		void shouldAddToCollection() {
			map.put(KEY, List.of("value"));

			FormDataMappingUtils.addToMap(map, KEY, "value1");

			assertThat(map).containsEntry("key", List.of("value", "value1"));
		}

		@Test
		void shouldAddMap() {
			var map2 = Map.of("key2", "value2");

			FormDataMappingUtils.addToMap(map, KEY, map2);

			assertThat(map).containsEntry("key", map2);
		}
	}

	@Nested
	@DisplayName("Build GrpcFormField")
	class TestToGrpcFormField {

		private static String NAME = "name";
		private static String LABEL = "label value";
		private static String VALUE = "value";

		@Test
		void shouldSetValue() {
			var field = buildFormField();

			assertThat(field.getValue()).isEqualTo(GrpcFormFieldTestFactory.VALUE);
		}

		@Test
		void shouldMapName() {
			var field = buildFormField();

			assertThat(field.getName()).isEqualTo(GrpcFormFieldTestFactory.NAME);
		}

		@Test
		void shouldMapLabel() {
			var field = buildFormField();

			assertThat(field.getLabel()).isEqualTo(LABEL);
		}

		@Test
		void shouldMapWithoutLabel() {
			var field = FormDataMappingUtils.buildFormField(NAME, VALUE);

			assertThat(field).isEqualTo(GrpcFormFieldTestFactory.createBuilder().clearLabel().build());
		}

		private GrpcFormField buildFormField() {
			return FormDataMappingUtils.buildFormField(NAME, VALUE, LABEL);
		}

	}

	@Nested
	class TestGetLabel {

		@Test
		void shouldReturnLabel() {
			var label = FormDataMappingUtils.getLabel(Map.of(LABEL_KEY, FIELD_LABEL));

			assertThat(label).contains(FIELD_LABEL);
		}

		@Test
		void shouldHandleAtomicValue() {
			var label = FormDataMappingUtils.getLabel(FIELD_VALUE);

			assertThat(label).isEmpty();
		}

		@Test
		void shouldHandleMissingLabel() {
			var label = FormDataMappingUtils.getLabel(Map.of(VALUE_KEY, FIELD_VALUE));

			assertThat(label).isEmpty();
		}
	}
}